import click
from mpm.model import *
import pathlib
from mpm.config import ServerParamType
from columnar import columnar

@click.group(help='Add, remove, and synchronize Servers')
def server():
    pass

@server.command(help='Add a new server')
@click.pass_context
@click.argument('name')
@click.argument('path', type=click.Path(dir_okay=True, exists=True))
def add(ctx, name, path):
    ctx.obj['config'].add_server(name, path)
    ctx.obj['config'].save()
    click.echo("Added server {} in {}".format(name, path))

@server.command(help="Save a server's current plugin list to mpm.yaml")
@click.pass_context
@click.argument('servers', type=ServerParamType(), nargs=-1)
def lock(ctx, servers):
    config = ctx.obj['config']
    if len(servers) == 0:
        servers = config.servers()
    validRepos = []
    for repo in ctx.obj['config'].repositories():
        if not repo.exists():
            click.echo("Missing repository: '{}' not found at {}".format(repo.name, repo.path))
        else:
            validRepos.append(repo)
    for server in servers:
        click.echo('{} ({}):'.format(server.name, server.path))
        savedPlugins = []
        changedPlugins = []
        for state in server.pluginStates(validRepos):
            if isinstance(state, Installed):
                savedPlugins.append(state)
            elif isinstance(state, Outdated):
                changedPlugins.append(state)
            elif isinstance(state, MissingVersions):
                changedPlugins.append(state)
        rows = []
        for state in savedPlugins:
            newSpec = PluginSpec(state.plugin.name,
                                 state.currentVersion.version)
            if newSpec == state.plugin:
                continue
            rows.append([
                state.plugin.name,
                click.style('{} -> {}'.format(state.plugin.versionSpec,
                                  state.currentVersion.version), fg='green'),
                state.currentVersion.version
            ])
            server.set_plugin_version(state.plugin.name, state.currentVersion.version)
        for state in changedPlugins:
            if state.currentVersion is not None:
                rows.append([
                    click.style(state.plugin.name, fg='yellow'),
                    click.style('{} -> {}'.format(state.plugin.versionSpec,
                                      state.currentVersion.version),
                                fg='yellow'),
                    state.currentVersion.version
                ])
                server.set_plugin_version(state.plugin.name, state.currentVersion.version)
        if len(rows) > 0:
            click.echo(columnar(rows, ['Plugin', 'Wanted', 'Installed'],
                                no_borders=True))
        config.update_server(server.name, server.config)
    click.echo("Apply changes? [y/N]", nl=False)
    answer = click.getchar().lower()
    if answer == "y":
        config.save()

@server.command('list', help='List configured servers')
@click.pass_context
@click.argument('servers', type=ServerParamType(), nargs=-1)
def list_(ctx, servers):
    if len(servers) == 0:
        servers = ctx.obj['config'].servers()
    for server in servers:
        click.echo('{} ({}):'.format(server.name, server.path))
        serverPath = pathlib.Path(server.path)
        if not serverPath.exists():
            click.echo(click.style("\tServer path does not exist", fg='red'))
            continue
        pluginPath = pathlib.Path(server.path+'/plugins')
        if not pluginPath.exists():
            click.echo(click.style("\tServer plugin directory does not exist", fg='red'))
            continue
        installed = []
        outdated = []
        missing = []
        unmanaged = []
        for state in sorted(server.pluginStates(ctx.obj['config'].repositories())):
            if isinstance(state, Outdated):
                outdated.append(state)
            elif isinstance(state, Installed):
                installed.append(state)
            elif isinstance(state, MissingVersions):
                missing.append(state)
            elif isinstance(state, UnmanagedFile):
                unmanaged.append(state)

        rows = []
        for state in sorted(installed):
            rows.append([
                state.plugin.name,
                str(state.plugin.versionSpec),
                str(state.currentVersion.version),
                str(state.currentVersion.version),
            ])
        for state in sorted(outdated):
            if state.currentVersion is None:
                curStr = click.style('Missing', fg='red')
            else:
                curStr = click.style(str(state.currentVersion.version), fg='yellow')
            rows.append([
                state.plugin.name,
                click.style(str(state.plugin.versionSpec), fg='green'),
                curStr,
                click.style(str(state.wantedVersion), fg='green')
            ])
        for state in sorted(missing):
            rows.append([
                click.style(state.plugin.name, fg='red'),
                click.style(str(state.plugin.versionSpec), fg='yellow'),
                click.style('Missing', fg='red'),
                click.style('Missing', fg='red')
            ])
        for state in sorted(unmanaged):
            rows.append([
                click.style(state.filename, fg='yellow'),
                '?',
                click.style('Unmanaged', fg='yellow', bold=True),
                ''
            ])

        if len(rows) > 0:
            click.echo(columnar(rows, ['Plugin', 'Wanted', 'Installed',
                                       'Available'], no_borders=True))
        else:
            click.echo(click.style('No plugins configured!', fg='yellow'))

@server.command(help='Add a plugin to a server')
@click.pass_context
@click.argument('server', type=ServerParamType())
@click.argument('plugins', nargs=-1)
def add_plugin(ctx, server, plugins):
    pluginQueue = []
    for pluginSpec in plugins:
        if os.path.exists(pluginSpec):
            plugin = Plugin(pluginSpec)
            pluginSpec = PluginSpec(plugin.name, str(plugin.version))
        else:
            allVersions = []
            for repo in ctx.obj['config'].repositories():
                allVersions += repo.versionsForPlugin(pluginSpec)
            pluginSpec = PluginSpec(pluginSpec, list(reversed(sorted(allVersions)))[0])

        pluginQueue.append(pluginSpec)

    click.echo("Resolved plugin list:")
    for pluginSpec in pluginQueue:
        click.echo("\t{} {}".format(click.style(pluginSpec.name, fg='green'),
                                    click.style(pluginSpec.versionSpec,
                                    fg='yellow')))
    click.echo("Add these plugins to server {}? [y/N] ".format(server.name),
               nl=False)
    answer = click.getchar().lower()
    if answer == "y":
        for pluginSpec in pluginQueue:
            try:
                server.add_plugin(pluginSpec)
            except KeyError:
                click.echo("{} already added!".format(pluginSpec))
        ctx.obj['config'].update_server(server.name, server.config)
        ctx.obj['config'].save()
        click.echo("Added!")
    else:
        click.echo("Cancelled.")

@server.command(help="Synchronize a server's plugins")
@click.pass_context
@click.argument('servers', type=ServerParamType(), nargs=-1)
def sync(ctx, servers):
    allServers = ctx.obj['config'].servers()
    if len(servers) > 0:
        allServers = servers
    validRepos = []
    for repo in ctx.obj['config'].repositories():
        if not repo.exists():
            click.echo("Missing repository: '{}' not found at {}".format(repo.name, repo.path))
        else:
            validRepos.append(repo)
    for server in allServers:
        if not server.exists():
            click.echo("Server '{}' not found at {}".format(server.name,
                                                            server.path))
            continue
        click.echo('{} ({}):'.format(server.name, server.path))
        outdated = []
        missing = []
        for state in sorted(server.pluginStates(validRepos)):
            if isinstance(state, Outdated):
                outdated.append(state)
            elif isinstance(state, MissingVersions):
                missing.append(state)

        rows = []

        for state in sorted(outdated):
            if state.currentVersion is None:
                curStr = click.style('Missing', fg='red')
            else:
                curStr = click.style(str(state.currentVersion.version), fg='yellow')
            rows.append([
                state.plugin.name,
                click.style(str(state.plugin.versionSpec), fg='green'),
                curStr,
                click.style(str(state.wantedVersion.version), fg='green')
            ])

        click.echo("Missing plugins:")
        for state in sorted(missing):
            rows.append([
                click.style(state.plugin.name, fg='red'),
                click.style(str(state.plugin.versionSpec), fg='yellow'),
                click.style('Missing', fg='red'),
                click.style('Missing', fg='red')
            ])

        if len(rows) > 0:
            click.echo(columnar(rows, ['Plugin', 'Wanted', 'Installed',
                                       'Available'], no_borders=True))

        if len(outdated):
            click.echo("Apply changes? [y/N]", nl=False)
            answer = click.getchar().lower()
            if answer == "y":
                with click.progressbar(length=len(outdated),
                                       label='Synchronizing') as bar:
                    i = 0
                    for state in outdated:
                        server.syncToVersion(state.wantedVersion)
                        i += 1
                        bar.update(i)
            else:
                click.echo("Not applying changes.")
        else:
            click.echo("No changes to apply.")
