import os
import yaml
import click
from mpm.repo import Repo
from mpm.server import Server
from mpm.model import *

try:
        from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
        from yaml import Loader, Dumper

DEFAULT_CONFIG = """
repositories: {}
servers: {}
"""

class RepoParamType(click.ParamType):
    name = "repo-name"

    def convert(self, value, param, ctx):
        if isinstance(value, Repo):
            return value
        return ctx.obj['config'].repository(value)

class ServerParamType(click.ParamType):
    name = "server-name"

    def convert(self, value, param, ctx):
        if isinstance(value, Repo):
            return value
        return ctx.obj['config'].server(value)

class Config():

    def __init__(self, path):
        if path is None:
            path = os.path.expanduser('~/mpm.yaml')
        self.path = path
        with open(path, 'r') as fd:
            self.yaml = yaml.load(fd, Loader=Loader)
        self.config = yaml.load(DEFAULT_CONFIG, Loader=Loader)
        if isinstance(self.yaml, dict):
            self.config.update(self.yaml)

    def repository(self, name):
        return Repo(name, self.config['repositories'][name])

    def repositories(self):
        return [Repo(name, c) for (name, c) in self.config['repositories'].items()]

    def update_repository(self, name, config):
        self.config['repositories'][name] = config

    def add_repository(self, name, path):
        if name in self.config['repositories']:
            raise ValueError('Repository already exists')

        self.update_repository(name, {
            'path': str(path)
        })

    def template(self, name):
        return self.config['templates'].get(name, dict())
    
    def mergeConfig(self, serverConf):
        baseConf = serverConf
        baseConf['plugins'] = []
        mergedPlugins = {}
        templates = [self.template(n) for n in serverConf['inherit']]
        for template in templates:
            for plugin in template.get('plugins', []):
                if plugin['name'] not in mergedPlugins:
                    mergedPlugins[plugin['name']] = dict()
                mergedPlugins[plugin['name']].update(plugin)
        for plugin in serverConf['plugins']:
            if plugin['name'] not in mergedPlugins:
                mergedPlugins[plugin['name']] = dict()
            mergedPlugins[plugin['name']].update(plugin)
        for plugin in mergedPlugins:
            baseConf['plugins'].append(mergedPlugins[plugin])
        print(baseConf)
        return baseConf

    def servers(self):
        return [Server(name, self.mergeConfig(c)) for (name, c) in self.config['servers'].items()]

    def server(self, name):
        return Server(name, self.mergeConfig(self.config['servers'][name]))

    def update_server(self, server, config):
        self.config['servers'][server] = config

    def add_server(self, name, path):
        if name in self.config['servers']:
            raise ValueError("Server already exists")

        self.update_server(name, {
            'path': path,
            'plugins': [],
            'inherit': []
        })

    def save(self):
        stream = open(self.path, 'w')
        yaml.dump(self.config, stream, Dumper=Dumper)
        stream.close()
