from semantic_version import SimpleSpec, Version
from functools import total_ordering
import os
import re

version_pattern = re.compile('^(?P<name>.+)-(?P<version>(?:\.?\d+)+).+jar$')

@total_ordering
class Plugin:
    def __init__(self, path, forceName=None, forceVersion=None):
        self.path = path
        pluginName = os.path.basename(path)

        if forceVersion is None:
            pluginMatches = version_pattern.match(pluginName)

            if pluginMatches is None:
                raise ValueError("Cannot derive plugin name from '{}'".format(path))

            if forceName is None:
                self.name = pluginMatches['name']
            else:
                self.name = forceName

            try:
                self.version = Version.coerce(pluginMatches['version'])
            except ValueError:
                raise ValueError("Cannot derive semver from '{}'".format(path))
        else:
            self.version = forceVersion
            self.name = forceName

    def normalizedFilename(self):
        return "{}-{}.jar".format(self.name, self.version)

    def unlink(self):
        os.unlink(self.path)

    def __repr__(self):
        return "Plugin({}, {} {})".format(self.path, self.name, self.version)

    def __eq__(self, other):
        return self.name == other.name and self.version == other.version

    def __ne__(self, other):
        return self.name != other.name or self.version != other.version

    def __lt__(self, other):
        if self.name == other.name:
            return self.version < other.version
        return self.name < other.name

@total_ordering
class PluginSpec:
    def __init__(self, name, versionSpec):
        self.name = name
        try:
            self.versionSpec = SimpleSpec(str(versionSpec))
        except ValueError:
            raise ValueError("Invalid version spec for plugin {}: {}".format(name, versionSpec))

    def __repr__(self):
        return "PluginSpec({}, {})".format(self.name, self.versionSpec)

    def __str__(self):
        return "{} {}".format(self.name, self.versionSpec)

    def __eq__(self, other):
        return self.name == other.name and self.versionSpec == other.versionSpec

    def __ne__(self, other):
        return self.name != other.name or self.versionSpec != other.versionSpec

    def __lt__(self, other):
        return self.name < other.name

@total_ordering
class PluginState:
    def __init__(self, plugin):
        self.plugin = plugin

    def __eq__(self, other):
        return self.plugin == other.plugin

    def __ne__(self, other):
        return self.plugin != other.plugin

    def __lt__(self, other):
        return self.plugin.name < other.plugin.name

class UnmanagedFile(PluginState):
    def __init__(self, filename):
        self.filename = filename

    def __lt__(self, other):
        if isinstance(other, UnmanagedFile):
            return self.filename < other.filename
        return False

class Outdated(PluginState):
    def __init__(self, plugin, currentVersion, wantedVersion):
        super().__init__(plugin)
        self.currentVersion = currentVersion
        self.wantedVersion = wantedVersion

    def __str__(self):
        return 'Outdated({})'.format(self.plugin)

class MissingVersions(PluginState):
    def __init__(self, plugin, currentVersion):
        super().__init__(plugin)
        self.currentVersion = currentVersion

class Installed(PluginState):
    def __init__(self, plugin, currentVersion):
        super().__init__(plugin)
        self.currentVersion = currentVersion

    def __str__(self):
        return 'Installed({} @ )'.format(self.plugin, self.currentVersion)
